/* NVTV i810 TV-I2C access -- Dirk Thierbach <dthierbach@gmx.de>
 *
 * This file is part of nvtv, a tool for tv-output on NVidia cards.
 * 
 * nvtv is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * nvtv is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id$
 *
 * Contents:
 *
 * Routines to control the TV and the I2C bus on Intel i810 etc. cards.
 *
 */

#include "debug.h"
#include "error.h"
#include "xfree.h"
#include "mmio.h"
#include "bitmask.h"
#include "tv_i810.h"
#include "xf86i2c.h"
#include "tv_null.h"
#include "tv_bt.h"
#include "tv_cx.h"
#include "tv_ch.h"
#include "tv_ph.h"

/* -------- Registers -------- */

#define I810_WR32(p,i,d)  MMIO_OUT32(((p)->MMIOBase), (i), (d))
#define I810_RD32(p,i)    MMIO_IN32(((p)->MMIOBase), (i))

#define I810_GPIOA	0x05010
#define I810_GPIOB	0x05014

#define I810_DCLK(i)	(0x06000 + (i)*4)
#define I810_DCLK_0D	0x06000
#define I810_DCLK_1D	0x06004
#define I810_DCLK_2D	0x06008
#define I810_LCD_CLKD	0x0600c
#define	I810_DCLK_0DS	0x06010
#define	I810_PWR_CLKC	0x06014

#define	I810_LCDTV_C 	0x60018
#define	I810_OVRACT  	0x6001c
#define	I810_BCLRPAT 	0x60020
#define	I810_LCD_CLKD	0x0600c

/* -------- I2C Driver Routines -------- */

/* bit locations in the register */

#define I2C_SCL_IN    (1 << 4)
#define I2C_SDA_IN    (1 << 12)

static void I810TvI2CGetBits(I2CBusPtr b, int *clock, int *data)
{
  I810Ptr pI810 = I810PTR(xf86Screens[b->scrnIndex]);
  register CARD32 val;

  I810_WR32(pI810, I810_GPIOB, 0x0);
  val = I810_RD32(pI810, I810_GPIOB);

  *clock = (val & I2C_SCL_IN) != 0;
  *data  = (val & I2C_SDA_IN) != 0;

  DEBUG(ErrorF("I810TvI2CGetBits(%p) val=0x%x, returns clock %d, data %d\n",
	       b, val, *clock, *data));
}

static void I810TvI2CPutBits(I2CBusPtr b, int clock, int data)
{
  I810Ptr pI810 = I810PTR(xf86Screens[b->scrnIndex]);

  if (clock) {
    I810_WR32(pI810, I810_GPIOB, 0x5);
    I810_WR32(pI810, I810_GPIOB, 0x4);
  } else { 
    I810_WR32(pI810, I810_GPIOB, 0x7);
    I810_WR32(pI810, I810_GPIOB, 0x6);
  }
  
  if (data) {
    I810_WR32(pI810, I810_GPIOB, 0x500);
    I810_WR32(pI810, I810_GPIOB, 0x400);
  } else { 
    I810_WR32(pI810, I810_GPIOB, 0x700);
    I810_WR32(pI810, I810_GPIOB, 0x600);
  }
  
  DEBUG(ErrorF("I810TvI2CPutBits(%p, %d, %d) val=0x%x\n", 
	       b, clock, data, val));
}

I2CBusPtr I810TvI2CInit (ScrnInfoPtr pScrn, char *name, int private)
{
  I2CBusPtr I2CPtr;

  I2CPtr = xf86CreateI2CBusRec();
  if (I2CPtr) {
    I2CPtr->BusName    = name;
    I2CPtr->scrnIndex  = pScrn->scrnIndex;
    I2CPtr->I2CPutBits = I810TvI2CPutBits;
    I2CPtr->I2CGetBits = I810TvI2CGetBits;
    /* Not sure about the delays. Matthew's code has HOLD_TIME = 100, and
     * uses that for all timings. The values used are about 100x larger 
     * than the default, so it should work, hopefully.
     */
    /* FIXME: Make these shorter, and see if it still works */
    I2CPtr->AcknTimeout  = 500; /* usec, default: 5 */
    I2CPtr->HoldTime     = 100; /* usec, default: 5 */
    I2CPtr->BitTimeout   = 500; /* usec, default: 5 */
    I2CPtr->ByteTimeout  = 500; /* usec, default: 5 */
    I2CPtr->AcknTimeout  = 500; /* usec, default: 5 */
    I2CPtr->StartTimeout = 500; /* usec, default: 5 */
    I2CPtr->RiseFallTime = 100; /* usec, default: 2 */
    I2CPtr->DriverPrivate.val = private; 
  }
  return I2CPtr;
}

Bool I810TvBusInit (ScrnInfoPtr pScrn)
{
  Bool ok;
  I810Ptr pI810 = I810PTR(pScrn);

  pI810->TvChain = NULL;
  pI810->TvDev = NULL;
  /* Must be named TV0, TV1, etc.; otherwise I2C_ID won't work */
  pI810->TvBus = I810TvI2CInit (pScrn, "TV0", 0);
  
  ok = TRUE;
  if (!xf86I2CBusInit(pI810->TvBus)) {
    ErrorF ("Cannot init bus\n");
    ok = FALSE;
  }
  return ok;
}

/* -------- CRT -------- */

void I810SetTvRegs (I810Ptr pI810, TVI810Regs *r)
{
  I810_WR32 (pI810, 0x60000, SetBitField (r->tvHDisplay,10:0,10:0) 
                           | SetBitField (r->tvHTotal,11:0,27:16));
  I810_WR32 (pI810, 0x60004, SetBitField (r->tvHBlankStart,11:0,11:0) 
                           | SetBitField (r->tvHBlankEnd,11:0,27:16));
  I810_WR32 (pI810, 0x60008, SetBitField (r->tvHSyncStart,11:0,11:0)
                           | SetBitField (r->tvHSyncEnd,11:0,27:16));
  I810_WR32 (pI810, 0x6000C, SetBitField (r->tvVDisplay,10:0,10:0)
                           | SetBitField (r->tvVTotal,11:0,27:16));
  I810_WR32 (pI810, 0x60010, SetBitField (r->tvVBlankStart,11:0,11:0)
                           | SetBitField (r->tvVBlankEnd,11:0,27:16));
  I810_WR32 (pI810, 0x60014, SetBitField (r->tvVSyncStart,11:0,11:0)
                           | SetBitField (r->tvVSyncEnd,11:0,27:16));
}

void I810GetTvRegs (I810Ptr pI810, TVI810Regs *r)
{
  r->tvHDisplay    = SetBitField (I810_RD32 (pI810, 0x60000),10: 0,10:0); 
  r->tvHTotal      = SetBitField (I810_RD32 (pI810, 0x60000),27:16,11:0);
  r->tvHBlankStart = SetBitField (I810_RD32 (pI810, 0x60004),11: 0,11:0);
  r->tvHBlankEnd   = SetBitField (I810_RD32 (pI810, 0x60004),27:16,11:0);
  r->tvHSyncStart  = SetBitField (I810_RD32 (pI810, 0x60008),11: 0,11:0);
  r->tvHSyncEnd    = SetBitField (I810_RD32 (pI810, 0x60008),27:16,11:0);
  r->tvVDisplay    = SetBitField (I810_RD32 (pI810, 0x6000C),10: 0,10:0);
  r->tvVTotal      = SetBitField (I810_RD32 (pI810, 0x6000C),27:16,11:0);
  r->tvVBlankStart = SetBitField (I810_RD32 (pI810, 0x60010),11: 0,11:0);
  r->tvVBlankEnd   = SetBitField (I810_RD32 (pI810, 0x60010),27:16,11:0);
  r->tvVSyncStart  = SetBitField (I810_RD32 (pI810, 0x60014),11: 0,11:0);
  r->tvVSyncEnd    = SetBitField (I810_RD32 (pI810, 0x60014),27:16,11:0);
}

void I810SetClock (I810Ptr pI810, int index, int m, int n, int p, int vco)
{
  register CARD32 val;

  I810_WR32 (pI810, I810_DCLK(index), SetBitField (n,9:0,25:16)
	                            | SetBitField (m,9:0,9:0));
  /* FIXME for index 3, bit 0 of lcdtv_c must be zero */
  val = I810_RD32 (pI810, I810_DCLK_0DS);
  val &= ~ (0x7c << index * 8);
  val |= (p & 0x7) << (index * 8 + 4) | (vco & 0x1) << (index * 8 + 2);
  I810_WR32 (pI810, I810_DCLK_0DS, val);
}

/* -------- Device enable/disable -------- */

int I810GetDevFlags (I810Ptr pI810)
{
  if (GetBit (I810_RD32(pI810, I810_LCDTV_C), 31)) {
    /* FIXME: may also be flat panel */
    return DEV_TELEVISION;
  } else {
    return DEV_MONITOR;
  }
}

/* -------- Device routines -------- */

void I810SetTvDevice (I810Ptr pI810, I2CChainPtr chain, Bool init)
{
  DPRINTF ("i810 set dev %p %i\n", chain, init);
  TVSetTvEncoder (&pI810->tvEncoder, chain);
  tvBusOk = TRUE;
  pI810->tvEncoder.InitRegs (&pI810->tvEncoder, PORT_I810);
}

/*
 * Free chain and device records of all tv chips
 */

void I810DestroyDevices (I810Ptr pI810)
{
  if (!pI810) return;
  if (pI810->TvChain) TVDestroyChain (pI810->TvChain);
  pI810->TvChain = NULL;
  pI810->TvDev = NULL;
  TVDestroyDevices (&pI810->TvBus, 1);
}

/*
 * Free chain, device records, and all busses
 */

void I810DestroyBusses (I810Ptr pI810)
{
  if (!pI810) return;
  I810DestroyDevices (pI810);
  TVDestroyBusses (&pI810->TvBus, 1);
}

void I810ProbeTvDevices (I810Ptr pI810)
{
  if (!pI810) return;
  I810UpdateTvState (pI810);
  tvState = TV_OFF;
  TVProbeCreateKnown (&pI810->TvBus, 1, &pI810->TvChain); 
  pI810->TvDev = pI810->TvChain ? pI810->TvChain->dev : NULL;
}

/* -------- -------- */

void I810UpdateTvState (I810Ptr pI810)
{
#ifndef FAKE_MMIO
  if (GetBit (I810_RD32(pI810, I810_LCDTV_C), 31)) {
    tvState = TV_ON;
  } else {
    if (tvState != TV_BARS) tvState = TV_OFF;
  }
#endif
}

void I810SetTestImage (I810Ptr pI810, TVEncoderRegs *r)
{
  /* FIXME chip registers ... */
  I810UpdateTvState (pI810);
  if (tvState != TV_ON) {
    /* Only do bars if not in tv-mode */
    pI810->tvEncoder.SetState (&pI810->tvEncoder, r, TV_BARS);
  }
}

void I810SetTvMode (I810Ptr pI810, TVRegs *r)
{
  TVState oldState, newState; /* hack for now, FIXME */

  I810UpdateTvState (pI810);
  oldState = tvState;
  newState = (r->devFlags & DEV_TELEVISION) ? TV_ON : TV_OFF;
  DPRINTF ("i810 set tv mode %i -> %i\n", oldState, newState);

  if (oldState == TV_ON && newState != TV_ON) 
  {
    I810_WR32 (pI810, I810_PWR_CLKC, 0x103);
    I810_WR32 (pI810, I810_LCDTV_C, 0x20000006); /* turn off tv */
  }
  if (oldState != newState) {
    pI810->tvEncoder.SetState (&pI810->tvEncoder, &r->enc, newState);
  }
  if (newState != TV_OFF) 
  {
    pI810->tvEncoder.SetPort (&pI810->tvEncoder, r->portEnc);
    pI810->tvEncoder.SetRegs (&pI810->tvEncoder, &r->enc, newState);
  }
  if (oldState != TV_ON && newState == TV_ON) {
    /* border enable flag in lcdtv_c */
    I810SetTvRegs (pI810, &r->crtc.i810);
    I810SetClock (pI810, 3, 10, 1, 3, 0); /* m=10 n=1 p=3 vco=0 -> f=1 */
    I810_WR32 (pI810, I810_PWR_CLKC, 0x102);
    I810_WR32 (pI810, I810_LCDTV_C, 0xa0000007); /* turn on tv, turn off crt */
  }
  tvState = newState;
}

long I810GetTvStatus (I810Ptr pI810, int index)
{
  long result; 

  I810UpdateTvState (pI810);
  tvBusOk = TRUE;
  result = pI810->tvEncoder.GetStatus (&pI810->tvEncoder, index);
  return result;
}

TVConnect I810GetTvConnect (I810Ptr pI810)
{
  TVConnect result;

  I810UpdateTvState (pI810);
  tvBusOk = TRUE;
  result = pI810->tvEncoder.GetConnect (&pI810->tvEncoder);
  return result;
}

#if 0

{
  register long l;

  /* LCD/TV clock: m=10, n=1, p=3 (/8) just reflects clock from DVO */
  /* Intel PLL: Fclk = 4 * (Fref * (m+2)/(n+2)) >> p */
  /* I810CalcVCLK in xfree/.../driver/i810_driver.c */

  [0x60018/4] = ; /* lcdtv_c control */
  /* 0xa0000007 -> turn on tv, turn off crt */
  /* 0x20000006 -> turn off tv */
  /* vsync, hsync active low. blank# active high */
  /* bit7: border enable */

  [0x6001c/4]  /* ovract  overlay active start/end   -- not used */
  [0x60020/4]  /* bclrpat   border color */

  [0x600c/4] = n << 16 | m; /* lcd_clkd */
  l = [0x6010/4]; /* dclk_0ds */
  l &= ~0x7c000000;
  l |= p << 28 | 0 << 26; /* vco=0 -> divide by 4*m */
  [0x6010/4] = l; /* dclk_0ds */
  [0x6014/4] = ?; /* pwr_clkc */
  /* 0x102; -> turn on tv, turn off crt */
  /* 0x103; -> if crt_enable */
}

#endif

/*

06=C0/3F cfrb=1 m/s*=1
0D=80/0F only

0E=0B/F0 00=60/00 01=80/3F 02=6C/00 04=45/30 05=2E/00 06=C0/3F 07=64/00 
08=18/03 0C=33/00 0D=80/0F 0E=40/3F 0F=22/C0 10=00/0F 11=00/07 12=00/00 
13=40/18 14=0D/00 15=14/00 16=38/00 17=40/3F 18=13/00 19=A0/00 1A=20/00 
1B=0A/F0 1C=0E/C0 1D=07/F0 1E=0C/F0 1F=04/F0 20=10/EF 21=00/FE 2A=40/80 
2B=10/00 2C=84/00 2D=84/00 2E=BB/00 2F=9B/00 30=64/00 31=0A/00 32=EE/00 
33=26/80 34=2A/80 35=7E/00 36=9A/00 37=3F/80 38=E0/00 39=06/80 3A=04/00 
3B=FE/00 3C=7E/00 3D=00/F8 0A=19/00 08=00/FD 0B=30/00 08=01/FE 

*/

/*

DCLK0 25.175 MHz n=3 m=19   p= /16  vco= /4m
DCLK1 28.322 MHz n=16 m=83  p= /16  vco= /4m

*/
