/* NVTV Philips TV-I2C access -- Dirk Thierbach <dthierbach@gmx.de>
 *
 * This file is part of nvtv, a tool for tv-output on NVidia cards.
 * 
 * nvtv is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * nvtv is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: tv_ph.c,v 1.12 2002/04/18 06:02:42 dthierbach Exp $
 *
 * Contents:
 *
 * Routines to access the Brooktree chip registers via the I2C bus.
 *
 */

#include "xf86i2c.h"
#include "xfree.h"
#include "debug.h"
#include "error.h"
#include "bitmask.h"
#include "tv_i2c.h"
#include "tv_ph.h"

/* -------- Philips -------- */

typedef struct {
  int blckl_diff; /* not sure if it is really a difference, but we have */
  int blnnl_diff; /* to modify these levels somehow */
  int reg01; /* mask c0   */
  int reg03; /* mask c0   */
  int reg04; /* mask c0   */
  int reg05; /* mask 00   */
  int reg06; /* mask 00   */
  int reg07; /* mask 00   */
  int reg08; /* mask fc   */
  int reg10; /* mask c0   */
  int reg11; /* mask 00   */
  int reg12; /* mask 00   */
  int reg14; /* mask 00   */
  int reg15; /* mask 80   */
  int reg20; /* mask 40 ! */
  int reg21; /* mask 00   */
  int reg24; /* mask 00   */
  int reg25; /* mask 80   */
  int reg30; /* mask c0   */
  int reg31; /* mask f8   */
  int reg33; /* mask 00   */
  int reg36; /* mask 00   */
  int reg60; /* mask 80   */
  int reg7d; /* mask 16 ! */ 
} TVPhMacro;

TVPhMacro tvPhNtscMacro0 = {
  blckl_diff:0, blnnl_diff:0, reg7d:0x00 
};

TVPhMacro tvPhNtscMacro1 = {         
  blckl_diff: 31, /* 0x1b <- 0x3a */ 
  blnnl_diff: 32, /* 0x0e <- 0x2e */ 
  reg60:0x40, reg7d:0xc8, reg01:0x1e,  
  reg03:0x11, reg04:0x26, reg05:0x0a, reg06:0xc5, reg07:0xff, reg08:0x03,  
  reg10:0x11, reg11:0xad, reg12:0x60, reg14:0x0f, reg15:0x78, reg20:0x9b,  
  reg21:0x04, reg24:0x00, reg25:0x00, reg30:0x1b, reg31:0x04, reg33:0x0f,  
  reg36:0x0f 
};                                                          

TVPhMacro tvPhNtscMacro2 = {                                             
  blckl_diff: 31, /* 0x1b <- 0x3a */ 
  blnnl_diff: 32, /* 0x0e <- 0x2e */ 
  reg60:0x40, reg7d:0xe8, reg01:0x1e,  
  reg03:0x11, reg04:0x26, reg05:0x0a, reg06:0xc5, reg07:0xff, reg08:0x03,  
  reg10:0x11, reg11:0xad, reg12:0x60, reg14:0x0f, reg15:0x78, reg20:0x9b,  
  reg21:0x04, reg24:0x00, reg25:0x00, reg30:0x1b, reg31:0x04, reg33:0x0f,  
  reg36:0x0f 
};                                                          

TVPhMacro tvPhNtscMacro3 = {                                             
  blckl_diff: 31, /* 0x1b <- 0x3a */ 
  blnnl_diff: 32, /* 0x0e <- 0x2e */ 
  reg60:0x40, reg7d:0xe8, reg01:0x18,  
  reg03:0x15, reg04:0x22, reg05:0x0a, reg06:0xc5, reg07:0xff, reg08:0x03,  
  reg10:0x15, reg11:0xad, reg12:0x60, reg14:0x0f, reg15:0x78, reg20:0x9b,  
  reg21:0x04, reg24:0x00, reg25:0x00, reg30:0x1b, reg31:0x04, reg33:0x0f,  
  reg36:0x0f 
};                                                          

/* PAL diff: blckl: -0x1C  blnnl: -0x1C */

TVPhMacro tvPhPalMacro0 = {                                              
  blckl_diff:0, blnnl_diff:0, reg7d:0x00 
};                                  

TVPhMacro tvPhPalMacro1= {                                               
  blckl_diff: 28, /* 0x17 <- 0x33 */ 
  blnnl_diff: 28, /* 0x19 <- 0x35 */ 
  reg60:0x40, reg7d:0xc8, reg01:0x1b,  
  reg03:0x22, reg04:0x2b, reg05:0x08, reg06:0x74, reg07:0x55, reg08:0x01,  
  reg10:0x22, reg11:0x15, reg12:0x60, reg14:0x07, reg15:0x7e, reg20:0xbb,  
  reg21:0x72, reg24:0x02, reg25:0x54, reg30:0x2c, reg31:0x04, reg33:0xfe,  
  reg36:0x7e 
}; 

TVPhMacro *TVMacroPhRegs (TVEncoderRegs *r)
{
  DPRINTF ("tv macro ph %i\n", r->ph.macro); 
  switch (r->ph.macro) {
    case 0: /* NTSC APS 0 */
      return &tvPhNtscMacro0;
      break;
    case 1: /* NTSC APS 1 */
      return &tvPhNtscMacro1;
      break;
    case 2: /* NTSC APS 2 */
      return &tvPhNtscMacro2;
      break;
    case 3: /* NTSC APS 3 */
      return &tvPhNtscMacro3;
      break;
    case 4: /* PAL APS 0 */
      return &tvPhPalMacro0;
      break;
    case 5: /* PAL APS 1 */
    case 6:
    case 7:
      return &tvPhPalMacro1;
      break;
  }
  return NULL;
}

void TVPhCreate (TVEncoderObj *this, TVChip chip, I2CDevPtr dev)
{
  DPRINTF ("tv create ph\n");
  this->chip = chip;
  this->dev = dev;
  this->hwconfig = 0; /* No hardware config for PH */
  this->hwstate = 0;  /* pcbn=0, slave=0 */
}

/*
  Registers:
  ----
  0x16:  Common DAC adjust fine - set gain to 0
  0x17:  Red DAC adjust fine - default
  0x18:  Green DAC adjust fine - default
  0x19:  Blue DAC adjust fine - default (max)
  0x1a:  Monitor sense mode - recommended setting
  0x1b:  Monitor sense mode off
  0x21:  Undocumented register in SAA7108E
  0x26:  Widescreen signal
  0x27:  Widescreen signal
  0x28:  Realtime control, burst start
  0x29:  Sync reset enable, burst end
  0x2a:  copy gen 0
  0x2b:  copy gen 1
  0x2c:  copy gen 2, Copy gen enable (off)
  0x2d:  Output port control - default connector
  0x38:  Gain luminance for RGB
  0x39:  Gain color difference for RGB
  0x3a:  Input control 1 - 0
  0x54:  Input cont. 2 (VPS enable) - edge1: falling, edge2: rising
  0x6c:  Horiz. trigger
  0x6d:  Horiz. trigger
  0x6e:  Multi control
  0x6f:  Closed caption, teletext - ttxen=0 ccen=01 sccln=2
  0x75:  Composite sync advance - 22h
  96-7   efs   = 0  FSVGC (FIELD) ignored
  96-6   pcbn  = 0  NCBO (BLANK) polarity
  96-5   slave = 0
  97-7   hfs   = 0  no hsync from FSVGC (FIELD)
  97-6   vfs   = 0  no vsync from FSVGC (FIELD)
  97-5   ofs   = 1
  97-4   pfs   = 0  fsync polarity (0:high, 1:low)
  97-3   ovs   = 1  vsync direction (0:in, 1:out)
  97-2   pvs   = 1  vsync polarity (0:high, 1:low)  (why 1 ???)
  97-1   ohs   = 1  hsync direction (0:in, 1:out)
  97-0   phs   = 0  hsync polarity (0:high, 1:low)
         Note: sync polarity varies wildly in different modes.
  0xa2:  Border color Y
  0xa3:  Border color U
  0xa4:  Border color V
  0xfc:  Vertical hot spot
  fd-7   lutoff = 1  (bypass LUT)
  fd-6   cmode  = 0  (cursor mode: invert)
  fd-5   lutl   = 0  (LUT loading via input inactive)
  fd-4:2 if     = 110 = 6 (R8G8B8 4:4:4 non-interlaced, special ordering)
  fd-1   matoff = 0  (RGB -> CrYCb active)
  fd-0   dfoff  = 0  (4:4:4 -> 4:4:2 active)
*/

void TVPhSetPort (TVEncoderObj *this, int port)
{
  DPRINTF ("tv port ph %08X\n", port);

  if ((port & PORT_FORMAT_MASK) != PORT_FORMAT_RGB) {
    ERROR ("TVPhSetPort: Only RGB multiplex format supported.\n");
    exit (1);
  }
  if ((port & PORT_PCLK_MODE) != PORT_PCLK_MASTER) {
    ERROR ("TVPhSetPort: Only clock master mode supported.\n");
    exit (1);
  }
  if ((port & PORT_PCLK_POLARITY) != PORT_PCLK_HIGH) {
    ERROR ("TVPhSetPort: Only active high clock supported.\n");
    exit (1);
  }
  if ((port & PORT_BLANK_DIR) != PORT_BLANK_OUT) {
    ERROR ("TVPhSetPort: Only output blank supported.\n");
    exit (1);
  }
  this->hwstate = ((port & PORT_BLANK_POLARITY) ? SetBit(6) : 0)
                | ((port & PORT_SYNC_DIR) ? SetBit(5) : 0);
  /* FIXME PCLK polarity in hwstate ? */
  TVWriteBus (this->dev, 0x96, SetBitField(this->hwstate,6:5,6:5));
  /* FIXME Test sync polarity - this varies wildly in the observed modes. */
  TVWriteBus (this->dev, 0x97, 0x20 /* ofs=1, hfs=vfs=0 */
	      | ((port & PORT_SYNC_DIR) ? 0 : SetBit(1) | SetBit(3))
	      | ((port & PORT_VSYNC_POLARITY) ? 0 : SetBit(2))
	      | ((port & PORT_HSYNC_POLARITY) ? 0 : SetBit(0)));
 
}

void TVPhInitRegs (TVEncoderObj *this, int port)
{
  register I2CDevPtr dev = this->dev;
  int i;

  DPRINTF ("nv init ph %04X\n", this->chip);
#if 0 /* TEST_PHILIPS */
  if ((this->chip & TV_PHILIPS_MODEL) == TV_PHILIPS_MODEL2) 
    { ERROR ("SAA7104 not yet supported\n"); exit (1); }
#endif
  TVPhSetPort (this, port);
  for (i = 0x01; i <= 0x16; i++) {
    TVWriteBus (dev, i, 0x00);  /* NV: 0x01 - 0x16 */ 
  }
  TVWriteBus (dev, 0x17, 0x1b); /* NV */ 
  TVWriteBus (dev, 0x18, 0x1b); /* NV */ 
  TVWriteBus (dev, 0x19, 0x1f); /* NV */ 
  TVWriteBus (dev, 0x1a, 0x46); /* NV */ 
  TVWriteBus (dev, 0x1b, 0x00); /* NV - monitor sense mode off */
  TVWriteBus (dev, 0x21, 0x1b); /* NV does it */
  TVWriteBus (dev, 0x26, 0xff); /* NV */ 
  TVWriteBus (dev, 0x27, 0x3f); /* NV */ 
  TVWriteBus (dev, 0x28, 0x19); /* NV */
  TVWriteBus (dev, 0x29, 0x1d); /* NV */
  TVWriteBus (dev, 0x2a, 0xff); /* NV - copy gen */
  TVWriteBus (dev, 0x2b, 0xff); /* NV - copy gen */
  TVWriteBus (dev, 0x2c, 0x0f); /* NV - copy gen off */
  TVWriteBus (dev, 0x2d, 0xb4); /* NV - default connector */
  if ((this->chip & TV_PHILIPS_MODEL) == TV_PHILIPS_MODEL2) {
    for (i = 0x2e; i <= 0x36; i++) {
      TVWriteBus (dev, i, 0x00);   /* NV: 0x2e - 0x36*/ 
    }
    TVWriteBus (dev, 0x37, 0x12);  /* NV: 0x37= 0x12 */ 
  } else {
    for (i = 0x2e; i <= 0x37; i++) {
      TVWriteBus (dev, i, 0x00);   /* NV: 0x2e - 0x37 */ 
    }
  }
  TVWriteBus (dev, 0x38, 0x1a); /* NV */
  TVWriteBus (dev, 0x39, 0x1a); /* NV */
  TVWriteBus (dev, 0x3a, 0x00); /* NV */ 
  if ((this->chip & TV_PHILIPS_MODEL) == TV_PHILIPS_MODEL2) {
    TVWriteBus (dev, 0x54, 0x02); /* NV - edge1: falling, edge2: rising */
  } else {
    TVWriteBus (dev, 0x54, 0x01); /* NV - edge1: falling, edge2: rising */
  }
  TVWriteBus (dev, 0x55, 0x0f); /* NV */
  TVWriteBus (dev, 0x56, 0xc3); /* NV */
  TVWriteBus (dev, 0x57, 0x00); /* NV */
  TVWriteBus (dev, 0x58, 0x02); /* NV */
  TVWriteBus (dev, 0x59, 0x30); /* NV */
  TVWriteBus (dev, 0x6c, 0x01); /* NV */
  TVWriteBus (dev, 0x6d, 0x20); /* NV */
  TVWriteBus (dev, 0x6f, 0x42); /* NV ttxen=0 ccen=01 sccln=2 */
  TVWriteBus (dev, 0x75, 0xb0); /* NV - csynca */ 
  TVWriteBus (dev, 0xa2, 0x10); /* NV */
  TVWriteBus (dev, 0xa3, 0x80); /* NV */
  TVWriteBus (dev, 0xa4, 0x80); /* NV */
  TVWriteBus (dev, 0xfc, 0x03); /* NV */ 
  TVWriteBus (dev, 0xfd, 0x98); /* NV */ 
}

/*
  Registers:
  ----
  0x28: Starting point of burst in clock cycles
  0x29: Sync reset enable, burst end
  0x2d: Output port control
  0x38: Gain luminance for RGB
  0x39: Gain color difference for RGB
  0x5a: Chrominance phase
  0x5b: Gain U
  0x5c: Gain V
  0x5d: Gain U MSB, black level
  0x5e: Gain V MSB, blanking level
  0x5f: Blanking level, CCR
  0x61: Standard control
  0x62: Burst amplitude
  0x63: Subcarrier 0
  0x64: Subcarrier 1
  0x65: Subcarrier 2
  0x66: Subcarrier 3
  0x6E: Multi Control (Field length FLC, Carrier reset PHRES, ...)
  0x70: Active display window horizontal start
  0x71: Active display window horizontal end
  0x72: Active display window horizontal values MSB
  0x7a: First active line
  0x7b: Last active line
  0x7c: MSBs for fal and lal, teletext mode
  0x81: Pixel clock 0
  0x82: Pixel clock 1
  0x83: Pixel clock 2
  0x90: Horizontal offset
  0x91: Horizontal pixel count
  0x92: Vertical offset (odd)
  0x93: Vertical offset (even)
  0x94: MSBs for pcl, pixel number and vertical offsets
  0x95: Line number (vertical pixel count)
  0x96: MSB for line number, Scaler control
  0x98: Line length
  0x99: Line lenth MSB, Input delay
  0x9a: Horizontal increment
  0x9b: Vertical increment
  0x9c: MSBs for horizontal and vertical increments
  0x9d: Weighting factor (odd)
  0x9e: Weighting factor (even)
  0x9f: Weighting factor MSBs
  0xa0: Vertical line skip
  0xa1: Vertical line MSB, blank enable
  0xa2: Border color luminance
  0xa3: Border color chrominance red
  0xa4: Border color chrominance blue
*/

void TVPhSetRegs (TVEncoderObj *this, TVEncoderRegs *r, TVState state)
{
  register I2CDevPtr dev = this->dev;
  TVPhMacro *m;
  int blckl, blnnl;
  int edge;

  DPRINTF ("nv set ph %04X\n", this->chip);
#if 0 /* TEST_PHILIPS */
  if ((this->chip & TV_PHILIPS_MODEL) == TV_PHILIPS_MODEL2) 
    { ERROR ("SAA7104 not yet supported\n"); exit (1); }
#endif
  m = TVMacroPhRegs (r);
  blckl = r->ph.blckl;
  blnnl = r->ph.blnnl;
  if (m) {
    blckl -= m->blckl_diff;
    blnnl -= m->blnnl_diff;
    if (blckl < 0) blckl = 0;
    if (blnnl < 0) blnnl = 0;
  }
  TVWriteBus (dev, 0x28, SetBitField(r->ph.bs,5:0,5:0));
  TVWriteBus (dev, 0x29, SetBitField(r->ph.be,5:0,5:0)
 	               | 0x00 /* sres=0 ttx pin */ );
  TVWriteBus (dev, 0x2d, SetBitField(r->ph.flags2,7:4,7:4)
 	               | 0x04 /* clk2en=1, encoff=0 */ );
  TVWriteBus (dev, 0x38, SetBitField(r->ph.gy,4:0,4:0));
  TVWriteBus (dev, 0x39, SetBitField(r->ph.gcd,4:0,4:0));

  edge = 0x01; /* edge1: falling, edge2: rising */
  if ((this->chip & TV_PHILIPS_MODEL) == TV_PHILIPS_MODEL2) {
     edge ^= 0x03;
  }
  if (r->ph.flags1 & PH_FLAG1_EDGE) 
  {
     edge ^= 0x03;
  }
  TVWriteBus (dev, 0x54, edge); 
  TVWriteBus (dev, 0x5a, Set8Bits(r->ph.chps));
  TVWriteBus (dev, 0x5b, Set8Bits(r->ph.gainu));
  TVWriteBus (dev, 0x5c, Set8Bits(r->ph.gainv));
  TVWriteBus (dev, 0x5d, SetBitField(blckl,5:0,5:0)
 	               | SetBitField(r->ph.gainu,8:8,7:7));
  TVWriteBus (dev, 0x5e, SetBitField(blnnl,5:0,5:0)
 	               | SetBitField(r->ph.gainv,8:8,7:7));
  TVWriteBus (dev, 0x5f, SetBitField(r->ph.blnnl,5:0,5:0) /* blnvb */
                       | SetBitField(r->ph.ccrs,1:0,7:6));
  TVWriteBus (dev, 0x61, Set8Bits(r->ph.flags1 & PH_FLAG1_MASK));

  /* FIXME: State for DOWN flags !! Always up at the moment */
  /* #define PH_FLAG1_DOWNA (1 << 6)
     #define PH_FLAG1_DOWND (1 << 7) */
  TVWriteBus (dev, 0x62, SetBitField(r->ph.bsta,6:0,6:0));
  TVWriteBus (dev, 0x63, SetBitField(r->ph.fsc,7:0,7:0));
  TVWriteBus (dev, 0x64, SetBitField(r->ph.fsc,15:8,7:0));
  TVWriteBus (dev, 0x65, SetBitField(r->ph.fsc,23:16,7:0));
  TVWriteBus (dev, 0x66, SetBitField(r->ph.fsc,31:24,7:0));
  /* FIXME: Is this really needed? */
  if ((this->chip & TV_PHILIPS_MODEL) == TV_PHILIPS_MODEL2) {
    TVWriteBus (dev, 0x6e, 0x00);
  } else {
    TVWriteBus (dev, 0x6e, SetBitField(r->ph.phres,1:0,5:4)
                         | SetBitField(r->ph.flc,1:0,1:0)
 	                 | 0x00 /* blckon, ldel are zero */ );
  }
  TVWriteBus (dev, 0x70, Set8Bits(r->ph.adwhs));
  TVWriteBus (dev, 0x71, Set8Bits(r->ph.adwhe));
  TVWriteBus (dev, 0x72, SetBitField(r->ph.adwhs,10:8,2:0)
                       | SetBitField(r->ph.adwhe,10:8,6:4));
  TVWriteBus (dev, 0x7a, Set8Bits(r->ph.fal));
  TVWriteBus (dev, 0x7b, Set8Bits(r->ph.lal));
  TVWriteBus (dev, 0x7c, SetBitField(r->ph.fal,8:8,4:4)
                       | SetBitField(r->ph.lal,8:8,6:6)
 	               | 0x00 /* TTX */ );
  TVWriteBus (dev, 0x81, SetBitField(r->ph.pcl,7:0,7:0));
  TVWriteBus (dev, 0x82, SetBitField(r->ph.pcl,15:8,7:0));
  TVWriteBus (dev, 0x83, SetBitField(r->ph.pcl,23:16,7:0));
  if ((this->chip & TV_PHILIPS_MODEL) == TV_PHILIPS_MODEL2) {
    if (r->ph.flags3 & PH_FLAG3_DOUBLE) {
      TVWriteBus (dev, 0x84, 0xe0); /* Divider E:I = 1:1 (doubles) */
    } else {
      TVWriteBus (dev, 0x84, 0xe5); /* Divider E:I = 2:2 (normal) */
    }
    TVWriteBus (dev, 0x85, 0x88);
  }
  TVWriteBus (dev, 0x90, Set8Bits(r->ph.xofs));
  TVWriteBus (dev, 0x91, Set8Bits(r->ph.xpix));
  TVWriteBus (dev, 0x92, Set8Bits(r->ph.yofso));
  TVWriteBus (dev, 0x93, Set8Bits(r->ph.yofse));
  TVWriteBus (dev, 0x94, SetBitField(r->ph.yofse,9:8,7:6)
                       | SetBitField(r->ph.yofso,9:8,5:4)
                       | SetBitField(r->ph.xpix,9:8,3:2)
	               | SetBitField(r->ph.xofs,9:8,1:0));
  TVWriteBus (dev, 0x95, Set8Bits(r->ph.ypix));
  TVWriteBus (dev, 0x96, SetBitField(r->ph.ypix,9:8,1:0)
 	               | SetBitFlag(r->ph.flags1,PH_FLAG1_YFIL,3)
 	               | SetBitField(this->hwstate,6:5,6:5)
	               | 0x00 /* rest all zero */ );
  TVWriteBus (dev, 0x98, Set8Bits(r->ph.hlen));
  /* hlen extra bit for SAA7105, 0 in SAA7102 */
  TVWriteBus (dev, 0x99, SetBitField(r->ph.hlen,11:8,3:0)
	               | SetBitField(r->ph.idel,3:0,7:4));
  TVWriteBus (dev, 0x9a, Set8Bits(r->ph.xinc));
  TVWriteBus (dev, 0x9b, Set8Bits(r->ph.yinc));
  TVWriteBus (dev, 0x9c, SetBitField(r->ph.xinc,11:8,3:0)
	               | SetBitField(r->ph.yinc,11:8,7:4));
  TVWriteBus (dev, 0x9d, Set8Bits(r->ph.yiwgto));
  TVWriteBus (dev, 0x9e, Set8Bits(r->ph.yiwgte));
  TVWriteBus (dev, 0x9f, SetBitField(r->ph.yiwgto,11:8,3:0)
	               | SetBitField(r->ph.yiwgte,11:8,7:4));
  TVWriteBus (dev, 0xa0, Set8Bits(r->ph.yskip));
  TVWriteBus (dev, 0xa1, SetBitField(r->ph.yskip,11:8,3:0)
 	               | 0x00 /* blen=0 */ );
  TVWriteBus (dev, 0xa2, Set8Bits(r->ph.bcy));
  TVWriteBus (dev, 0xa3, Set8Bits(r->ph.bcu));
  TVWriteBus (dev, 0xa4, Set8Bits(r->ph.bcv));

  if (m) {
    TVWriteBus (dev, 0x60, SetBitField(m->reg60,7:0,7:0));
    TVWriteBus (dev, 0x7d, Set8Bits(m->reg7d & ~0x16));
    TVWriteBus (dev, 0x01, SetBitField(m->reg01,5:0,5:0));
    TVWriteBus (dev, 0x03, SetBitField(m->reg03,5:0,5:0));
    TVWriteBus (dev, 0x04, SetBitField(m->reg04,5:0,5:0));
    TVWriteBus (dev, 0x05, SetBitField(m->reg05,7:0,7:0));
    TVWriteBus (dev, 0x06, SetBitField(m->reg06,7:0,7:0));
    TVWriteBus (dev, 0x07, SetBitField(m->reg07,7:0,7:0));
    TVWriteBus (dev, 0x08, SetBitField(m->reg08,1:0,1:0));
    TVWriteBus (dev, 0x10, SetBitField(m->reg10,5:0,5:0));
    TVWriteBus (dev, 0x11, SetBitField(m->reg11,7:0,7:0));
    TVWriteBus (dev, 0x12, SetBitField(m->reg12,7:0,7:0));
    TVWriteBus (dev, 0x14, SetBitField(m->reg14,7:0,7:0));
    TVWriteBus (dev, 0x15, SetBitField(m->reg15,6:0,6:0));
    TVWriteBus (dev, 0x20, Set8Bits(m->reg20 & ~0x40));
    TVWriteBus (dev, 0x21, SetBitField(m->reg21,7:0,7:0));
    TVWriteBus (dev, 0x24, SetBitField(m->reg24,7:0,7:0));
    TVWriteBus (dev, 0x25, SetBitField(m->reg25,6:0,6:0));
    TVWriteBus (dev, 0x30, SetBitField(m->reg30,5:0,5:0));
    TVWriteBus (dev, 0x31, SetBitField(m->reg31,3:0,3:0));
    TVWriteBus (dev, 0x33, SetBitField(m->reg33,7:0,7:0));
    TVWriteBus (dev, 0x36, SetBitField(m->reg36,7:0,7:0));
  }
}

void TVPhSetState (TVEncoderObj *this, TVEncoderRegs *r, TVState state)
{
  register I2CDevPtr dev = this->dev;

  /* FIXME */
  /* 0x6e blckon, ... */
  /* FIXME 0x61 dacoff/chip off ... but need connect state ... always on*/

  tvState = state;
  switch (state)
  {
    case TV_OFF: 
      TVWriteBus (dev, 0x3a, 0x80); /* NV */ 
      TVWriteBus (dev, 0x61, Set8Bits(r->ph.flags1 & PH_FLAG1_MASK));
      break;
    case TV_BARS: 
      TVWriteBus (dev, 0x3a, 0x80); /* NV + cbenb=1 */ 
      TVWriteBus (dev, 0x61, Set8Bits(r->ph.flags1 & PH_FLAG1_MASK));
      break;
    case TV_ON: 
      TVWriteBus (dev, 0x3a, 0x00); /* NV */ 
      TVWriteBus (dev, 0x61, Set8Bits(r->ph.flags1 & PH_FLAG1_MASK));
      break;
    default:
      break;
  }
}

long TVPhGetStatus (TVEncoderObj *this, int index)
{ 
  I2CByte fifo, status;

  TVReadBus (this->dev, 0x00, &status); /* FIFO status */
  TVReadBus (this->dev, 0x80, &fifo); /* FIFO status */
  
  return SetBitField (status,1:0,1:0) | SetBitField (fifo,1:0,3:2) |
    ((tvState == TV_ON) ? 0x100 : 0);
}

TVConnect TVPhGetConnect (TVEncoderObj *this)
{
  TVConnect status = CONNECT_BOTH;
  I2CByte msense;

  /* Enable monitor sensing temporarily */
  TVWriteBus (this->dev, 0x1a, 0x2c); /* NV */
  TVWriteBus (this->dev, 0x1b, 0x80);

  /* Read monitor sense data */
  TVReadBus (this->dev, 0x1b, &msense); /* Monitor sense */

  DPRINTF("TVPhGetConnect %02X\n", msense);

  switch (msense & 0x7) { /* 0=active, 1=inactive */
    case 0x06: 
      status = CONNECT_FBAS;
      break;
    case 0x01:
      status = CONNECT_SVHS;
      break;
    case 0x05:
      status = CONNECT_CONVERT;
      break;
    case 0x07:
      ERROR("Warning: No TV seems to be connected\n");
      /* no break */
    default:
      status = CONNECT_BOTH;
      break;
  }

  /* Disable monitor sensing */
  TVWriteBus (this->dev, 0x1b, 0x00 );

  return status;
}

/*
 * Check for Philips chip on device dev. Return version string if found, 
 * NULL otherwise.
 */

char *TVDetectPhilips (I2CDevPtr dev, TVChip *encoder)
{
  I2CByte cid, ver;
  char *name;
  static char version [50];

#ifdef FAKE_PROBE_PHILIPS
#ifdef FAKE_PROBE_ID
  *encoder = FAKE_PROBE_ID;
#else
  *encoder = TV_PHILIPS;
#endif
  snprintf (version, 50, "Philips Fake (%1s:%02X)", I2C_ID(dev));
  return version;
#else
#ifdef FAKE_I2C
  return NULL;
#endif
#endif

#if 0 //// FIXME REMOVE
  tvBusOk = TRUE;  
  TVWriteBus (dev, 0xff, 0x00);
  DPRINTF ("ph check LUT (%s)\n", tvBusOk ? "ok" : "err");
  if (!tvBusOk) return NULL;
  /* If we're here, it's not a BT, but may be an CX, in either read mode. */
  tvBusOk = TRUE;  
  TVWriteBus (dev, 0x01, 0x00); /* should be zero according to docs, anyway */
  DPRINTF ("ph check 0x01 (%s)\n", tvBusOk ? "ok" : "err");
  if (!tvBusOk) return NULL;
  TVReadBus (dev, 0x00, &ver); 
  TVReadBus (dev, 0x01, &res); 
  DPRINTF ("ph check %02X %02X (%s)\n", ver, res, tvBusOk ? "ok" : "err");
  if (res != 0x00 || ver == res) return NULL;
  /* Do it again, in case there were just changes it status */
  TVReadBus (dev, 0x00, &ver); 
  TVReadBus (dev, 0x01, &res); 
  DPRINTF ("ph check %02X %02X (%s)\n", ver, res, tvBusOk ? "ok" : "err");
  if (res != 0x00 || ver == res) return NULL;
#endif

  tvBusOk = TRUE;  
  TVReadBus (dev, 0x00, &ver); 
  TVReadBus (dev, 0x1c, &cid); 
  ver = (ver & 0xe0 ) >> 5; /* 3 highest bits are version info */
  DPRINTF ("ph check ver=%01x cid=%02x (%s)\n", ver, cid, tvBusOk?"ok":"err");
  *encoder = TV_PHILIPS;
  if (ver != 0x2 && ver != 0x5) {
    snprintf (version, 50, "Unknown Philips #%i (%1s:%02X)", ver, I2C_ID(dev));
    return version;
  }
  if (!tvBusOk) return NULL;
  /* Probe addr 0x40 and 0x42 to identify SAA7108, SAA7109 */
  if (xf86I2CProbeAddress(dev->pI2CBus, 0x40) || 
      xf86I2CProbeAddress(dev->pI2CBus, 0x42)) {
    switch (cid) {
      case 0x02: name = "7108";  *encoder = TV_PHILIPS_7108; break;
      case 0x03: name = "7109";  *encoder = TV_PHILIPS_7109; break;
      case 0x04: name = "7108A"; *encoder = TV_PHILIPS_7108A; break;
      case 0x05: name = "7109A"; *encoder = TV_PHILIPS_7109A; break;
      default:   name = NULL; break;
    }
  } else {
    switch (cid) {
      case 0x02: name = "7102"; *encoder = TV_PHILIPS_7102; break;
      case 0x03: name = "7103"; *encoder = TV_PHILIPS_7103; break;
      case 0x04: name = "7104"; *encoder = TV_PHILIPS_7104; break;
      case 0x05: name = "7105"; *encoder = TV_PHILIPS_7105; break;
      default:   name = NULL; break;
    }
  }
  if (name) {
    TVReadBus (dev, 0x00, &ver); 
    if (!tvBusOk) return NULL;
    snprintf (version, 50, "Philips SAA%s (%1s:%02X)", name, I2C_ID(dev));
    return version;
  } else {
    snprintf (version, 50, "Philips (id = %02X) (%1s:%02X)", 
	      cid, I2C_ID(dev));
    return version;
  }
}

#ifdef NVTV_DEBUG
int PhilipsRegs[] =
{
  0x00,
  0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c,
  0x22,
  0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b ,0x2c, 0x2d,
  0x38, 0x39, 0x3a,
  0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
  0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
  0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
  0x80, 0x81, 0x82, 0x83, 0x84,
  0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
  0xa0, 0xa1, 0xa2, 0xa3, 0xa4,
  0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff,
};

void PhilipsRegDump (I2CDevPtr dev)
{
  I2CByte value;
  int i,j;
  char regstring[50];
  char binary[32];
  
  DPRINTF("Philips register dump:\n" );
  DPRINTF("[Reg] [dec] [bin(7-0)]\n" );
  for( i = 0; i < ( sizeof(PhilipsRegs) / sizeof(PhilipsRegs[0])); i++ ) {
    TVReadBus (dev, PhilipsRegs[i], &value);
    
    /* Convert to binary */
    for(j=0;j < 8;j++) {
      if( ( value & ( 0x01 << j ) ) != 0 ) binary[7 - j] = '1';
      else binary[7 - j] = '0';
    }
    binary[8] = '\0';
    
    snprintf( regstring, 50, "0x%02x:  %03d   %s\n", PhilipsRegs[i], value, binary );
    DPRINTF( regstring );
  }
}
#endif /* NVTV_DEBUG */

TVEncoderObj tvPhTemplate = {
  chip: TV_PHILIPS, dev: NULL, minClock: 10000, maxClock: 44000, 
  Create:     TVPhCreate,
  InitRegs:   TVPhInitRegs, 
  SetRegs:    TVPhSetRegs, 
  SetPort:    TVPhSetPort,
  SetState:   TVPhSetState,
  GetConnect: TVPhGetConnect, 
  GetStatus:  TVPhGetStatus
};

/* SAA7104: 

maxClock 85 MHz 
PLL between ca. 40 Mhz and 85 Mhz (before divide)
PAL/NTSC up to 1280x1024 (1920x1080 interlaced)
HDTV up to 1920x1080 interlaced, 1280x720 progressive
I2C at 400kHz (also 7102)
MV rev 7.01, rev 6.1, rev 1.03 (525p)
Anti-Flicker 5 lines
2 FIFOs (one for upsampling)
seperate CSYNC pin (also 7102)

Different registers:

1b: 6=msa 5=msoe ; automatic monitor sense, ms pin TVD output enable
2d: 2=cvbsen2 ; pin RED... normal or CVBS
37: 6=yupsc 5=yfil1 4=yfil0 2=gzoom 1=igain 0=xint ; Input path control
    (nv=12: yfil=1 (before), igain=1)
3a: 5=syntv
54: 5=gpval 4=gpen 1=edge 0=slot
61: 5=inpi
62: 7=rtce
6e: 7=nvtrig ; vtrig sign
7c: 5=ttxo
80: 3=iferr 2=bferr ; input fifo overflow, buffer fifo overflow (for yupsc)
84: 7=dclk 6=pclsy 5=ifra 4=ifbp 3=pcle1 2=pcle0 1=pcli1 0=pcli0
    (nv=e_: dclk=single, pclsy=sync, ifra=reset, ifbp=active)
    (nv=_5: e:i=2:2, nv=_0: e:i=1:1)
85: 7=eidiv 3=fili3 2=fili2 1=fili1 0=fili0
    (nv=88: input dividers on, threshold=8)
96: 2=hsl is n/a
99: 3=hlen11
d0-dc: hdtv stuff

*/

